/**
 * DDS Function Generator
 *
 * Kerry D. Wong
 * http://www.kerrywong.com
 *
 * 11/2012
 */

#include <Keypad.h>
#include <LiquidCrystal.h>
#include <EEPROM.h>

//maximum allowed input length
const byte MAX_FREQ_DIGITS = 11;

const byte KEYPAD_ROWS = 4;
const byte KEYPAD_COLS = 3;

char keys[KEYPAD_ROWS][KEYPAD_COLS] = {
    {'1', '2', '3'},
    {'4', '5', '6'},
    {'7', '8', '9'},
    {'*', '0', '#'}
};

enum DisplayMode {
    DISP_MODE_DISPLAY,
    DISP_MODE_CHANGE_FREQUENCY
} dispMode = DISP_MODE_DISPLAY;

enum DisplayUnit {
    mHz, //Milli Hertz
    Hz, //Hertz
    KHz, //Killo Hertz 
    MHz, //Mega Hertz    
} dispUnit = Hz;

byte rowPins[KEYPAD_ROWS] = {5, 4, 3, 2};
byte colPins[KEYPAD_COLS] = {8, 7, 6};

//AD9850 pins
int PIN_RESET = 13;
int PIN_FREQ_UPDATE = 12;
int PIN_CLOCK = 11;
int PIN_DATA = 10;

//Multi-function button pin;
int PIN_BUTTON = 9;

double currentFreq, freqHz;
char inputBuf[MAX_FREQ_DIGITS]; //60,000,000,000
int bufPtr = 0;

boolean functionKeyPressed = false;

Keypad kpd = Keypad(makeKeymap(keys), rowPins, colPins, KEYPAD_ROWS, KEYPAD_COLS);
LiquidCrystal lcd(A0, A1, A2, A3, A4, A5);

void AD9850Reset() {
    digitalWrite(PIN_FREQ_UPDATE, 0);

    digitalWrite(PIN_RESET, 0);
    digitalWrite(PIN_RESET, 1);
    digitalWrite(PIN_RESET, 0);

    digitalWrite(PIN_CLOCK, HIGH);
    digitalWrite(PIN_CLOCK, LOW);

    digitalWrite(PIN_FREQ_UPDATE, 1);
    digitalWrite(PIN_FREQ_UPDATE, 0);
}

void AD9850Update(double frequency) {
    unsigned char i, w;
    long int tuningWord;

    tuningWord = frequency * 4294.967296f / 125.0f;

    shiftOut(PIN_DATA, PIN_CLOCK, LSBFIRST, tuningWord);
    shiftOut(PIN_DATA, PIN_CLOCK, LSBFIRST, tuningWord >> 8);
    shiftOut(PIN_DATA, PIN_CLOCK, LSBFIRST, tuningWord >> 16);
    shiftOut(PIN_DATA, PIN_CLOCK, LSBFIRST, tuningWord >> 24);
    shiftOut(PIN_DATA, PIN_CLOCK, LSBFIRST, 0x0);

    digitalWrite(PIN_FREQ_UPDATE, 1);
    digitalWrite(PIN_FREQ_UPDATE, 0);
}

void resetInputBuf() {
    for (int i = 0; i < MAX_FREQ_DIGITS; i++)
        inputBuf[i] = 0;

    bufPtr = 0;
}

void updateDisplay() {
    char u[5];

    if (dispUnit == mHz) strcpy(u, " mHz");
    else if (dispUnit == Hz) strcpy(u, " Hz");
    else if (dispUnit == KHz) strcpy(u, " KHz");
    else if (dispUnit == MHz) strcpy(u, " MHz");

    if (dispMode == DISP_MODE_DISPLAY) {
        lcd.setCursor(0, 0);
        for (int i = 0; i < 16; i++) lcd.print(" ");
        lcd.setCursor(0, 0);

        lcd.print(currentFreq);

        lcd.setCursor(12, 0);
        lcd.print(u);

        lcd.setCursor(0, 1);
        lcd.print("*Enter (F)Recall");
    } else { //set
        lcd.clear();
        lcd.setCursor(0, 1);
        lcd.print("*Cancel #Set");

        lcd.setCursor(0, 0);
        for (int i = 0; i < bufPtr; i++)
            lcd.print(inputBuf[i]);

        lcd.setCursor(12, 0);
        lcd.print(u);
    }
}

void setup() {
    lcd.begin(16, 2);

    pinMode(PIN_RESET, OUTPUT);
    pinMode(PIN_FREQ_UPDATE, OUTPUT);
    pinMode(PIN_CLOCK, OUTPUT);
    pinMode(PIN_DATA, OUTPUT);

    pinMode(PIN_BUTTON, INPUT);
    digitalWrite(PIN_BUTTON, HIGH);

    dispMode = DISP_MODE_DISPLAY;
    dispUnit = Hz;
    currentFreq = 1000;
    freqHz = currentFreq;

    AD9850Reset();
    AD9850Update(currentFreq);

    Serial.begin(9600);
    updateDisplay();
}

/**
 * Mode  Fn  *  #   [0-9]  action
 * Disp      x              enter set
 *       x   x              change unit
 *       x      x           reset
 *       x            x     save [0-9]  
 *                    x     recall [0-9]
 * Set       x              Cancel
 *       x   x              change unit
 *       x      x           reset
 *              x           set    
 */
void loop() {
    char t;

    if (digitalRead(PIN_BUTTON) == LOW && !functionKeyPressed) {
        delay(20);
        if (digitalRead(PIN_BUTTON == LOW))
            functionKeyPressed = true;
    }

    char key = kpd.getKey();

    if (key) {
        switch (key) {
            case '#':
                if (functionKeyPressed) {
                    setup();
                } else {
                    if (dispMode == DISP_MODE_CHANGE_FREQUENCY) {
                        for (int i = bufPtr - 1; i >= 0; i--) {
                            t = inputBuf[bufPtr - 1 - i];
                            inputBuf[bufPtr - 1 - i] = inputBuf[i];
                            inputBuf[i] = t;
                        }

                        currentFreq = atof(inputBuf);

                        if (dispUnit == mHz) freqHz = currentFreq / 1000.0;
                        else if (dispUnit == Hz) freqHz = currentFreq;
                        else if (dispUnit == KHz) freqHz = currentFreq * 1000.0;
                        else if (dispUnit == MHz) freqHz = currentFreq * 1000000.0;

                        AD9850Update(freqHz);

                        resetInputBuf();
                        dispMode = DISP_MODE_DISPLAY;
                    }
                }
                break;
            case '*': //mode
                if (functionKeyPressed) { // Alt + * change frequency unit if in frequency setting mode
                    if (dispUnit == mHz) dispUnit = Hz;
                    else if (dispUnit == Hz) dispUnit = KHz;
                    else if (dispUnit == KHz) dispUnit = MHz;
                    else if (dispUnit == MHz) dispUnit = mHz;
                    else (dispUnit = Hz);

                    if (dispMode == DISP_MODE_DISPLAY) {
                        if (dispUnit == mHz) currentFreq = freqHz * 1000.0;
                        else if (dispUnit == Hz) currentFreq = freqHz;
                        else if (dispUnit == KHz) currentFreq = freqHz / 1000.0;
                        else if (dispUnit == MHz) currentFreq = freqHz / 1000000.0;
                    }
                } else { //change display mode
                    dispMode = dispMode == DISP_MODE_DISPLAY ? DISP_MODE_CHANGE_FREQUENCY : DISP_MODE_DISPLAY;

                    if (dispMode == DISP_MODE_DISPLAY) resetInputBuf();
                }
                break;
            case '0': case '1': case '2': case '3': case '4':
            case '5': case '6': case '7': case '8': case '9':
                if (dispMode == DISP_MODE_DISPLAY) {
                    byte buf[8], t, u;
                    
                    t = (byte) key - 48;
                    
                    if (functionKeyPressed) { // save frequency setting
                        memcpy(buf, &currentFreq, 8);

                        for (int i = 0; i < 8; i++)
                            EEPROM.write(t * 10 + i, buf[i]);

                        EEPROM.write(t * 10 + 8, dispUnit);
                    } else {//load frequency setting
                        for (int i = 0; i < 8; i++)
                            buf[i] = EEPROM.read(t * 10 + i);

                        u = EEPROM.read(t * 10 + 8);

                        if (u < 4) {
                            dispUnit = (DisplayUnit) u;
                            memcpy(&currentFreq, buf, 8);
                        }

                        if (dispUnit == mHz) freqHz = currentFreq / 1000.0;
                        else if (dispUnit == Hz) freqHz = currentFreq;
                        else if (dispUnit == KHz) freqHz = currentFreq * 1000.0;
                        else if (dispUnit == MHz) freqHz = currentFreq * 1000000.0;

                        AD9850Update(freqHz);

                        Serial.println(currentFreq);
                    }
                } else {
                    if (dispMode == DISP_MODE_CHANGE_FREQUENCY) {
                        if (bufPtr < MAX_FREQ_DIGITS) {
                            inputBuf[bufPtr] = key;

                            bufPtr++;
                        }
                    }
                }
                break;
        }

        updateDisplay();
    }

    if (digitalRead(PIN_BUTTON) == HIGH && functionKeyPressed) {
        delay(20);
        if (digitalRead(PIN_BUTTON) == HIGH)
            functionKeyPressed = false;
    }
}
